#ifndef __ASMUTILS_H__
#define __ASMUTILS_H__

#include <types.h>

/*******************************
   Some x86 insn definitions
********************************/

#define X86_IN_IMM_INSN      0xe4
#define X86_OUT_IMM_INSN     0xe6
#define X86_IN_REG_INSN      0xec
#define X86_OUT_REG_INSN     0xee
#define X86_INT_INSN         0xcd


#define X86_MAX_INSN_LEN     15

/* group1 prefix */
#define X86_PREFIX_REP       0xf3
/* group2 prefix */	     
#define X86_PREFIX_CS        0x2e
#define X86_PREFIX_SS        0x36
#define X86_PREFIX_DS        0x3e
#define X86_PREFIX_ES        0x26
#define X86_PREFIX_FS        0x64
#define X86_PREFIX_GS        0x65

/* group3/4 prefix force 32 or 16 bits depending on execution mode */	     
#define X86_PREFIX_OP        0x66
#define X86_PREFIX_ADDR      0x67

/* some string opcodes */
#define X86_OPCODE_INS8      0x6c
#define X86_OPCODE_INS32     0x6d

#define X86_OPCODE_OUTS8     0x6e
#define X86_OPCODE_OUTS32    0x6f

#define X86_OPCODE_STOS8     0xaa
#define X86_OPCODE_STOS32    0xab

#define X86_OPCODE_LODS8     0xac
#define X86_OPCODE_LODS32    0xad

#define X86_OPCODE_MOVS8     0xa4
#define X86_OPCODE_MOVS32    0xa5

/* vmx opcodes */
#define X86_OPCODE_VMCALL    0xc1010f


/**************************************
 *          IA32 MSRs
 **************************************/

/*
** Read/Write standard MSR (ecx=index, edx,eax=value)
*/
#define read_msr32(index,val)               asm volatile ( "rdmsr" :"=a"((val)):"c"((index)) )
#define write_msr32(index,val)              asm volatile ( "wrmsr" ::"c"((index)),"a"((val)) )

#define read_msr64(index,h_val,l_val)       asm volatile ( "rdmsr" :"=d"((h_val)),"=a"((l_val)):"c"((index)) )
#define write_msr64(index,h_val,l_val)      asm volatile ( "wrmsr" ::"c"((index)),"d"((h_val)),"a"((l_val)) )

/*
** IA32_APIC_BASE_MSR
*/
typedef struct ia32_apic_base_msr_fields
{
   uint_t    rsrvd0:8;
   uint_t    bsp:1;       /* (1) processor is BSP */
   uint_t    rsrvd1:2;
   uint_t    g_enable:1;  /* (1) Global enable */
   uint_t    paddr:24;    /* APIC base physical address */
   uint_t    rsrvd2:28;

} __attribute__((packed)) ia32_apic_base_msr_fields_t;

typedef union ia32_apic_base_msr
{
   raw_msr_entry_t              _msr;
   raw64_t                      raw;
   ia32_apic_base_msr_fields_t  fields;

} __attribute__((packed)) ia32_apic_base_msr_t;

#define read_msr_ia32_apic_base(val)        read_msr64(0x1b,(val)._msr.edx,(val)._msr.eax)
#define write_msr_ia32_apic_base(val)       write_msr64(0x1b,(val)._msr.edx,(val)._msr.eax)

#define apic_disabled()                     ({ ia32_apic_base_msr_t apic; read_msr_ia32_apic_base(apic); apic.fields.g_enable; })

#define enable_apic()						\
   ({								\
      ia32_apic_base_msr_t apic;				\
      read_msr_ia32_apic_base(apic);				\
      apic.fields.g_enable = 1;					\
      write_msr_ia32_apic_base(apic);				\
   })


/*
** IA32_SYSENTER_CS, IA32_SYSENTER_EIP, IA32_SYSENTER_ESP
*/
#define IA32_SYSENTER_CS_MSR                0x174
#define IA32_SYSENTER_ESP_MSR               0x175
#define IA32_SYSENTER_EIP_MSR               0x176

#define read_msr_ia32_sysenter_cs(val)      read_msr32(IA32_SYSENTER_CS_MSR,val)
#define read_msr_ia32_sysenter_esp(val)     read_msr64(IA32_SYSENTER_ESP_MSR,(val).high,(val).low)
#define read_msr_ia32_sysenter_eip(val)     read_msr64(IA32_SYSENTER_EIP_MSR,(val).high,(val).low)

/*
** IA32_DEBUGCTL_MSR
*/
typedef struct ia32_debugctl_msr_fields
{
   uint_t    lbr:1;    /* last branch int/excp */
   uint_t    btf:1;    /* single step on branch */
   uint_t    r1:4;     /* reserved (0) */
   uint_t    tr:1;     /* trace msg enable */
   uint_t    bts:1;    /* branch trace store */
   uint_t    btint:1;  /* branch trace interrupt */
   uint_t    r2:23;    /* reserved (0) */
} __attribute__((packed)) ia32_debugctl_msr_fields_t;

typedef union ia32_debugctl_msr
{
   raw_msr_entry_t             _msr;
   raw64_t                     raw;
   ia32_debugctl_msr_fields_t  fields;

} __attribute__((packed)) ia32_debugctl_msr_t;

#define read_msr_ia32_debugctl(val)      read_msr64(0x1d9,(val)._msr.edx,(val)._msr.eax)

/*
** MSR Feature Control
*/
#define IA32_FEATURE_CONTROL_MSR_LOCK_BIT   0
#define IA32_FEATURE_CONTROL_MSR_VMX_BIT    2

typedef struct ia32_feature_control_msr
{
   uint32_t  raw;
} feat_ctl_msr_t;

#define read_msr_feature_ctrl(feat)       read_msr32(0x3a,(feat).raw)
#define write_msr_feature_ctrl(feat)      write_msr32(0x3a,(feat).raw)


/*
** DR7 debug register
*/
typedef struct debug_register_7_fields
{
   uint_t    l0:1;
   uint_t    g0:1;
   uint_t    l1:1;
   uint_t    g1:1;
   uint_t    l2:1;
   uint_t    g2:1;
   uint_t    l3:1;
   uint_t    g3:1;
   uint_t    le:1;
   uint_t    ge:1;
   uint_t    r1:3;   /* reserved */
   uint_t    gd:1;
   uint_t    r2:2;   /* reserved */
   uint_t    rw0:2;
   uint_t    len0:2;
   uint_t    rw1:2;
   uint_t    len1:2;
   uint_t    rw2:2;
   uint_t    len2:2;
   uint_t    rw3:2;
   uint_t    len3:2;

   uint32_t  upper;  /* 64 bit alignment */
} __attribute__((packed)) dr7_reg_fields_t;

typedef union debug_register_7
{
   raw64_t            raw;
   dr7_reg_fields_t   fields;

} __attribute__((packed)) dr7_reg_t;


/*
** CR0
*/
#define CR0_PE_BIT      0
#define CR0_MP_BIT      1
#define CR0_EM_BIT      2
#define CR0_TS_BIT      3
#define CR0_ET_BIT      4
#define CR0_NE_BIT      5
#define CR0_WP_BIT     16
#define CR0_AM_BIT     18
#define CR0_NW_BIT     29
#define CR0_CD_BIT     30
#define CR0_PG_BIT     31

#define CR0_PE         (1<<CR0_PE_BIT)
#define CR0_MP         (1<<CR0_MP_BIT)
#define CR0_EM         (1<<CR0_EM_BIT)
#define CR0_TS         (1<<CR0_TS_BIT)
#define CR0_ET         (1<<CR0_ET_BIT)
#define CR0_NE         (1<<CR0_NE_BIT)
#define CR0_WP         (1<<CR0_WP_BIT)
#define CR0_AM         (1<<CR0_AM_BIT)
#define CR0_NW         (1<<CR0_NW_BIT)
#define CR0_CD         (1<<CR0_CD_BIT)
#define CR0_PG         (1<<CR0_PG_BIT)


typedef struct control_register_0_fields
{
   uint_t    pe:1;   /* protected mode */
   uint_t    mp:1;   /* monitor copro */
   uint_t    em:1;   /* emulation */
   uint_t    ts:1;   /* task switch */
   uint_t    et:1;   /* ext type */
   uint_t    ne:1;   /* num error */
   uint_t    r1:10;  /* reserved */
   uint_t    wp:1;   /* write protect */
   uint_t    r2:1;   /* reserved */
   uint_t    am:1;   /* align mask */
   uint_t    r3:10;  /* reserved */
   uint_t    nw:1;   /* not write through */
   uint_t    cd:1;   /* cache disable */
   uint_t    pg:1;   /* paging */

   uint32_t  upper;  /* 64 bit alignment */
} __attribute__((packed)) cr0_reg_fields_t;

typedef union control_register_0
{
   raw64_t           raw;
   cr0_reg_fields_t  fields;

} __attribute__((packed)) cr0_reg_t;

#define get_cr0()                         ({uint32_t val; asm volatile ( "mov %%cr0, %%eax":"=a"(val)); val;})
#define set_cr0(cr0_reg)                  asm volatile ( "mov %%eax, %%cr0"::"a"((cr0_reg).raw.low) )
#define enable_paging()                   asm volatile ( "mov %%cr0, %%eax; or $0x80000000, %%eax; mov %%eax, %%cr0":::"eax" )

/*
** CR3
*/
typedef struct control_register_3_fields
{
   uint_t     r1:3;          /* reserved */
   uint_t     pwt:1;         /* write through (1), write back (0) */
   uint_t     pcd:1;         /* page dir could be cached (0) */
   uint_t     r2:7;          /* reserved */
   uint_t     base_addr:20;  /* physical base addr of the page directory (align 4KB) */

   uint32_t   base_addr_high;     /* top most bits of bas addr under 64bits */

} __attribute__((packed)) cr3_reg_fields_t;

typedef union control_register_3
{
   raw64_t           raw;
   cr3_reg_fields_t  fields;

} __attribute__((packed)) cr3_reg_t;

#define get_cr3()                         ({uint32_t val; asm volatile ( "mov %%cr3, %%eax":"=a"(val)); val;})
#define set_cr3(cr3_reg)                  asm volatile ( "mov %%eax, %%cr3"::"a"((cr3_reg).raw.low) )
#define get_pdbr(val)                     asm volatile ( "mov %%cr3, %%eax":"=a"(val) )
#define set_pdbr(val)                     asm volatile ( "mov %%eax, %%cr3"::"a"(val) )


/*
** CR4
*/
#define CR4_VME_BIT          0
#define CR4_PVI_BIT          1
#define CR4_TSD_BIT          2
#define CR4_DE_BIT           3
#define CR4_PSE_BIT          4
#define CR4_PAE_BIT          5
#define CR4_MCE_BIT          6
#define CR4_PGE_BIT          7
#define CR4_PCE_BIT          8
#define CR4_OSFXSR_BIT       9
#define CR4_OSXMMEXCPT_BIT  10
#define CR4_VMXE_BIT        13

typedef struct control_register_4_fields
{
   uint_t    vme:1;        /* virtual 8086 */
   uint_t    pvi:1;        /* pmode virtual int */
   uint_t    tsd:1;        /* time stamp disable */
   uint_t    de:1;         /* debug ext */
   uint_t    pse:1;        /* page sz ext */
   uint_t    pae:1;        /* phys addr ext */
   uint_t    mce:1;        /* machine check enable */
   uint_t    pge:1;        /* page global enable */
   uint_t    pce:1;        /* perf counter enable */
   uint_t    osfxsr:1;     /* fxsave fxstore */
   uint_t    osxmmexcpt:1; /* simd fpu excpt */
   uint_t    r1:2;         /* reserved */
   uint_t    vmxe:1;       /* vmx enable */
   uint_t    r2:18;        /* reserved */

   uint32_t  upper;        /* 64 bit alignment */
} __attribute__((packed)) cr4_reg_fields_t;

typedef union control_register_4
{
   raw64_t           raw;
   cr4_reg_fields_t  fields;

} __attribute__((packed)) cr4_reg_t;

#define get_cr4()                         ({uint32_t val; asm volatile ( "mov %%cr4, %%eax":"=a"(val)); val;})
#define set_cr4(cr4_reg)                  asm volatile ( "mov %%eax, %%cr4"::"a"((cr4_reg).raw.low) )
#define enable_vmx()                      asm volatile ( "mov %%cr4, %%eax; or $0x2000, %%eax; mov %%eax, %%cr4":::"eax" )
#define enable_vme()                      asm volatile ( "mov %%cr4, %%eax; or $0x1, %%eax; mov %%eax, %%cr4":::"eax" )
#define disable_vme()                     asm volatile ( "mov %%cr4, %%eax; and $0xfffffffe, %%eax; mov %%eax, %%cr4":::"eax" )


/*
** RFLAGS
*/
#define RFLAGS_CF_BIT         0
#define RFLAGS_PF_BIT         2
#define RFLAGS_AF_BIT         4
#define RFLAGS_ZF_BIT         6
#define RFLAGS_SF_BIT         7
#define RFLAGS_TF_BIT         8
#define RFLAGS_IF_BIT         9
#define RFLAGS_DF_BIT        10
#define RFLAGS_OF_BIT        11
#define RFLAGS_VIF_BIT       19

typedef struct cpu_flags_register_fields
{
   uint_t   cf:1;     /* carry */
   uint_t   r1:1;     /* (1) */
   uint_t   pf:1;     /* parity */
   uint_t   r2:1;     /* (0) */
   uint_t   af:1;     /* adjust */
   uint_t   r3:1;     /* (0) */
   uint_t   zf:1;     /* zero */
   uint_t   sf:1;     /* sign */
   uint_t   tf:1;     /* trap */
   uint_t   IF:1;     /* int */
   uint_t   df:1;     /* div */
   uint_t   of:1;     /* overflow */

   uint_t   iopl:2;   /* io pvl */
   uint_t   nt:1;     /* nested task */
   uint_t   r4:1;     /* (0) */
   uint_t   rf:1;     /* resume */
   uint_t   vm:1;     /* virtual 8086 */
   uint_t   ac:1;     /* align */
   uint_t   vif:1;    /* virtual IF */
   uint_t   vip:1;    /* virtual Interrupt Pending */
   uint_t   id:1;     /* identification */
   uint_t   r5:10;    /* (0) */
} __attribute__((packed)) rflags_reg_fields_t;

typedef union cpu_flags_register
{
   raw64_t              raw;
   rflags_reg_fields_t  fields;

} __attribute__((packed)) rflags_reg_t;


/*
** CPUID
*/
#define __cpuid(eax,ebx,ecx,edx)		\
   asm volatile (				\
      "push %%eax       \n"			\
      "push %%ebx       \n"			\
      "push %%ecx       \n"			\
      "push %%edx       \n"			\
      "mov  %0, %%eax   \n"			\
      "mov  %1, %%ebx   \n"			\
      "mov  %2, %%ecx   \n"			\
      "mov  %3, %%edx   \n"			\
      "cpuid            \n"			\
      "mov  %%eax, %0   \n"			\
      "mov  %%ebx, %1   \n"			\
      "mov  %%ecx, %2   \n"			\
      "mov  %%edx, %3   \n"			\
      "pop  %%edx       \n"			\
      "pop  %%ecx       \n"			\
      "pop  %%ebx       \n"			\
      "pop  %%eax       \n"			\
      ::"m"(eax),"m"(ebx),"m"(ecx),"m"(edx) )


#define cpuid(index,ecx,edx)                asm volatile ( "cpuid":"=c"((ecx)),"=d"((edx)):"a"((index)) )
#define cpuid4(index,eax,ebx,ecx,edx)       asm volatile ( "cpuid":"=a"(eax),"=b"(ebx),"=c"((ecx)),"=d"((edx)):"a"((index)) )


/*
** CPUID Features & Extended Features
*/
#define CPUID_EXT_FEAT_MWAIT_BIT    3
#define CPUID_EXT_FEAT_VMX_BIT      5

#define CPUID_EXT_FEAT_MWAIT       (1<<CPUID_EXT_FEAT_MWAIT_BIT)
#define CPUID_EXT_FEAT_VMX         (1<<CPUID_EXT_FEAT_VMX_BIT)

#define CPUID_FEAT_FPU_BIT          0
#define CPUID_FEAT_VME_BIT          1
#define CPUID_FEAT_DE_BIT           2
#define CPUID_FEAT_PSE_BIT          3
#define CPUID_FEAT_TSC_BIT          4
#define CPUID_FEAT_MSR_BIT          5
#define CPUID_FEAT_PAE_BIT          6
#define CPUID_FEAT_MCE_BIT          7
#define CPUID_FEAT_CX8_BIT          8
#define CPUID_FEAT_APIC_BIT         9
#define CPUID_FEAT_SEP_BIT         11
#define CPUID_FEAT_MTRR_BIT        12
#define CPUID_FEAT_PGE_BIT         13
#define CPUID_FEAT_MCA_BIT         14
#define CPUID_FEAT_CMOV_BIT        15
#define CPUID_FEAT_PAT_BIT         16
#define CPUID_FEAT_PSE36_BIT       17
#define CPUID_FEAT_PSN_BIT         18
#define CPUID_FEAT_CLFSH_BIT       19
#define CPUID_FEAT_DS_BIT          21
#define CPUID_FEAT_ACPI_BIT        22
#define CPUID_FEAT_MMX_BIT         23
#define CPUID_FEAT_FXSR_BIT        24
#define CPUID_FEAT_SSE_BIT         25
#define CPUID_FEAT_SSE2_BIT        26
#define CPUID_FEAT_SS_BIT          27
#define CPUID_FEAT_HTT_BIT         28
#define CPUID_FEAT_TM_BIT          29
#define CPUID_FEAT_PBE_BIT         31

#define CPUID_FEAT_FPU             (1<<CPUID_FEAT_FPU_BIT)
#define CPUID_FEAT_VME             (1<<CPUID_FEAT_VME_BIT)
#define CPUID_FEAT_DE              (1<<CPUID_FEAT_DE_BIT)
#define CPUID_FEAT_PSE             (1<<CPUID_FEAT_PSE_BIT)
#define CPUID_FEAT_TSC             (1<<CPUID_FEAT_TSC_BIT)
#define CPUID_FEAT_MSR             (1<<CPUID_FEAT_MSR_BIT)
#define CPUID_FEAT_PAE             (1<<CPUID_FEAT_PAE_BIT)
#define CPUID_FEAT_MCE             (1<<CPUID_FEAT_MCE_BIT)
#define CPUID_FEAT_CX8             (1<<CPUID_FEAT_CX8_BIT)
#define CPUID_FEAT_APIC            (1<<CPUID_FEAT_APIC_BIT)
#define CPUID_FEAT_SEP             (1<<CPUID_FEAT_SEP_BIT)
#define CPUID_FEAT_MTRR            (1<<CPUID_FEAT_MTRR_BIT)
#define CPUID_FEAT_PGE             (1<<CPUID_FEAT_PGE_BIT)
#define CPUID_FEAT_MCA             (1<<CPUID_FEAT_MCA_BIT)
#define CPUID_FEAT_CMOV            (1<<CPUID_FEAT_CMOV_BIT)
#define CPUID_FEAT_PAT             (1<<CPUID_FEAT_PAT_BIT)
#define CPUID_FEAT_PSE36           (1<<CPUID_FEAT_PSE36_BIT)
#define CPUID_FEAT_PSN             (1<<CPUID_FEAT_PSN_BIT)
#define CPUID_FEAT_CLFSH           (1<<CPUID_FEAT_CLFSH_BIT)
#define CPUID_FEAT_DS              (1<<CPUID_FEAT_DS_BIT)
#define CPUID_FEAT_ACPI            (1<<CPUID_FEAT_ACPI_BIT)
#define CPUID_FEAT_MMX             (1<<CPUID_FEAT_MMX_BIT)
#define CPUID_FEAT_FXSR            (1<<CPUID_FEAT_FXSR_BIT)
#define CPUID_FEAT_SSE             (1<<CPUID_FEAT_SSE_BIT)
#define CPUID_FEAT_SSE2            (1<<CPUID_FEAT_SSE2_BIT)
#define CPUID_FEAT_SS              (1<<CPUID_FEAT_SS_BIT)
#define CPUID_FEAT_HTT             (1<<CPUID_FEAT_HTT_BIT)
#define CPUID_FEAT_TM              (1<<CPUID_FEAT_TM_BIT)
#define CPUID_FEAT_PBE             (1<<CPUID_FEAT_PBE_BIT)

#define cpuid_features(eCx,eDx)    cpuid(0x1,eCx,eDx)
#define vmx_supported()            ({ uint32_t c,d; cpuid_features(c,d); c & CPUID_EXT_FEAT_VMX; })
#define vme_supported()            ({ uint32_t c,d; cpuid_features(c,d); d & CPUID_FEAT_VME; })
#define apic_supported()           ({ uint32_t c,d; cpuid_features(c,d); d & CPUID_FEAT_APIC; })

/*
** CPUID Extended Processor Features
*/
#define CPUID_INTEL_EXT_PROC_FEAT_NX_BIT     20
#define CPUID_INTEL_EXT_PROC_FEAT_EM64T_BIT  29
#define CPUID_INTEL_EXT_PROC_FEAT_NX         (1<<CPUID_INTEL_EXT_PROC_FEAT_NX_BIT)
#define CPUID_INTEL_EXT_PROC_FEAT_EM64T      (1<<CPUID_INTEL_EXT_PROC_FEAT_EM64T_BIT)

#define CPUID_AMD_EXT_PROC_FEAT_SVM_BIT       2
#define CPUID_AMD_EXT_PROC_FEAT_SVM           (1<<CPUID_AMD_EXT_PROC_FEAT_SVM_BIT)

#define cpuid_ext_proc_features(eCx,eDx)      cpuid(0x80000001,eCx,eDx)
#define svm_supported()                       ({ uint32_t c,d; cpuid_ext_proc_features(c,d); c & CPUID_AMD_EXT_PROC_FEAT_SVM; })


/*
** AMD CPUID SVM features
*/
#define CPUID_AMD_SVM_NPT_FEATURE_BIT             0
#define CPUID_AMD_SVM_LBR_FEATURE_BIT             1
#define CPUID_AMD_SVM_LOCK_FEATURE_BIT            2
#define CPUID_AMD_SVM_NRIP_FEATURE_BIT            3
#define CPUID_AMD_SVM_SSE_FEATURE_BIT             9

#define CPUID_AMD_SVM_NPT_FEATURE                 (1<<CPUID_AMD_SVM_NPT_FEATURE_BIT)
#define CPUID_AMD_SVM_LBR_FEATURE                 (1<<CPUID_AMD_SVM_LBR_FEATURE_BIT)
#define CPUID_AMD_SVM_LOCK_FEATURE                (1<<CPUID_AMD_SVM_LOCK_FEATURE_BIT)
#define CPUID_AMD_SVM_NRIP_FEATURE                (1<<CPUID_AMD_SVM_NRIP_FEATURE_BIT)
#define CPUID_AMD_SVM_SSE_FEATURE                 (1<<CPUID_AMD_SVM_SSE_FEATURE_BIT)

typedef union amd_svm_features
{
   struct
   {
      uint32_t   eax,ebx,ecx,edx;
   };

   struct
   {
      uint8_t    svm_rev;    /* SVM revision */
      uint_t     rsrvd0:24;
      uint32_t   asid_nr;    /* max number of ASID */
      uint32_t   rsrvd1;
      uint_t     npt:1;      /* Nested Page Tables */
      uint_t     lbr:1;      /* LBR virt */
      uint_t     lock:1;     /* SVM Lock */
      uint_t     nrip:1;     /* NRIP save */
      uint_t     rsrvd2:5;
      uint_t     sse3_5:1;   /* SSE3 and SSE5 disable */
      uint_t     rsrvd3:22;
   };

} __attribute__((packed)) amd_svm_feat_t;

#define cpuid_amd_svm_features(feat)          cpuid4(0x8000000a,(feat).eax,(feat).ebx,(feat).ecx,(feat).edx)
#define cpuid_check_amd_svm_feature(feature)  ({ amd_svm_feat_t feats; cpuid_amd_svm_features(feats); feats.edx & (feature); })

#define svm_lock_available()                  cpuid_check_amd_svm_feature( CPUID_AMD_SVM_LOCK_FEATURE )
#define svm_npt_available()                   cpuid_check_amd_svm_feature( CPUID_AMD_SVM_NPT_FEATURE )


/*
** RDTSC
*/
#define rdtsc()                              ({ raw64_t x; asm volatile ( "rdtsc":"=a"(x.low),"=d"(x.high) ); x.raw; })


/*
** RDMSR/WRMSR
*/
#define __read_msr(eax,ecx,edx)         \
   asm volatile (			\
      "push %%eax       \n"		\
      "push %%ecx       \n"		\
      "push %%edx       \n"		\
      "mov  %0, %%eax   \n"		\
      "mov  %1, %%ecx   \n"		\
      "mov  %2, %%edx   \n"		\
      "rdmsr            \n"		\
      "mov  %%eax, %0   \n"		\
      "mov  %%ecx, %1   \n"		\
      "mov  %%edx, %2   \n"		\
      "pop  %%edx       \n"		\
      "pop  %%ecx       \n"		\
      "pop  %%eax       \n"		\
      ::"m"(eax),"m"(ecx),"m"(edx) )

#define __write_msr(eax,ecx,edx)        \
   asm volatile (			\
      "push %%eax       \n"		\
      "push %%ecx       \n"		\
      "push %%edx       \n"		\
      "mov  %0, %%eax   \n"		\
      "mov  %1, %%ecx   \n"		\
      "mov  %2, %%edx   \n"		\
      "wrmsr            \n"		\
      "mov  %%eax, %0   \n"		\
      "mov  %%ecx, %1   \n"		\
      "mov  %%edx, %2   \n"		\
      "pop  %%edx       \n"		\
      "pop  %%ecx       \n"		\
      "pop  %%eax       \n"		\
      ::"m"(eax),"m"(ecx),"m"(edx) )


/*
** power of 2
*/
#define _2_power(pow)                 (1<<(pow))

/*
** Binary-Coded-Decimal (BCD)
**
** Decimal:     43           52
** BCD:      0100 0011    0101 0010
**
** each nibble (4 bits part) can not be greater than 1001 (decimal 9)
*/
#define bin2bcd(_vAl_)    ((((_vAl_)&0xff)/10)<<4 | (_vAl_)%10)
#define bcd2bin(_VaL_)    ((((_VaL_)>>4)&0xf)*10 + ((_VaL_)&0xf))


/*
** Play with flags
*/
#define SET_FLAG(var, flag)   ((var)|=(1<<(flag)))
#define DEL_FLAG(var, flag)   ((var)&=~(1<<(flag)))


/*
** Usefull macro to read/write to io ports
*/
#define out(data,port)  asm volatile("outb %%al,%%dx"::"a"(data),"d"(port))
#define in(port)        ({unsigned char data;asm volatile("inb %%dx,%%al":"=a"(data):"d"(port));data;})


/*
** Save/Restore flags
*/
#define EFLAGS_IF          (1<<9)
#define EFLAGS_VM          (1<<17)
#define EFLAGS_VIF         (1<<19)
#define EFLAGS_VIP         (1<<20)

#define clear_intel_flags()        asm volatile( "push $0 ; popf" );
#define get_intel_flags()          ({uint32_t val;save_intel_flags(val);val;})
#define save_intel_flags(flags)    asm volatile( "pushf;pop %0":"=m"(flags)::"memory" )
#define load_intel_flags(flags)    asm volatile( "push %0;popf"::"m"(flags):"memory" )


/*
** Manage Maskable Interrupts
*/
#define force_interrupts_on()     asm volatile( "sti;nop" )
#define force_interrupts_off()    asm volatile( "cli" )

#define atomic_on()               force_interrupts_off()
#define atomic_off()              force_interrupts_on()

#define halt()                    asm volatile( "hlt" )

#define amd_gif_on()              asm volatile ( "stgi" )
#define amd_gif_off()             asm volatile ( "clgi" )

/*
** Inhibite/Restore interrupt flag
*/
#define interrupts_enabled()          (get_intel_flags() & EFLAGS_IF)
#define disable_interrupts(flags)     ({save_intel_flags(flags);atomic_on();})
#define enable_interrupts(flags)      ({save_intel_flags(flags);atomic_off();})
#define restore_interrupts(flags)     load_intel_flags(flags)


/*
#define disable_interrupts(flags)   \
({save_intel_flags(flags);asm volatile( "cli" );kuart_printf("%s:%s:DIS\n", __FILE__, __FUNCTION__);})

#define restore_interrupts(flags)   \
({kuart_printf("%s:%s:ENA\n", __FILE__,__FUNCTION__);load_intel_flags(flags);})
*/


/*
** Debug helpers
*/
//#define breakpoint()              asm volatile("int $3")
#define breakpoint()              asm volatile("1: jmp 1b")

/*
** Invalidate TLB entry
*/
#define invalidate(addr)     asm volatile( "invlpg %0"::"m"(addr) )

/*
** Asm code to get/set some registers
*/
#define get_ebp()            ({uint32_t val;asm volatile( "mov %%ebp, %%eax":"=a"(val));val;})
#define set_ebp(val)         asm volatile( "mov %0, %%ebp"::"m"(val) );

#define get_esp()            ({uint32_t val;asm volatile( "mov %%esp, %%eax":"=a"(val));val;})
#define set_esp(val)         asm volatile( "mov %0, %%esp"::"m"(val) );

#define get_fs()             ({uint16_t val;asm volatile( "xor %%eax,%%eax ; mov %%fs, %%ax":"=a"(val));val;})
#define get_gs()             ({uint16_t val;asm volatile( "xor %%eax,%%eax ; mov %%gs, %%ax":"=a"(val));val;})

#define get_ebx()            ({uint32_t val;asm volatile( "":"=b"(val) ); val;})

#define get_cr2()            ({uint32_t addr;asm volatile( "mov %%cr2, %%eax":"=a"(addr));addr;})
#define set_cr2(v)           asm volatile ( "mov   %%eax, %%cr2"::"a"(v) )

#define get_gdtr(aLocation)  asm volatile ( "sgdtl %0"::"m"(aLocation) )
#define get_ldtr(aLocation)  asm volatile ( "sldtl %0"::"m"(aLocation) )
#define get_idtr(aLocation)  asm volatile ( "sidtl %0"::"m"(aLocation) )
#define get_tr(aLocation)    asm volatile ( "str   %0"::"m"(aLocation) )

#define set_gdtr(val)        asm volatile ( "lgdt  %0"::"m"(val) )
#define set_ldtr(val)        asm volatile ( "lldt  %0"::"m"(val) )
#define set_ldtr_reg(val)    asm volatile ( "lldt  %%ax"::"a"(val) )
#define set_idtr(val)        asm volatile ( "lidt  %0"::"m"(val) )
#define set_tr(val)          asm volatile ( "ltr   %%ax"::"a"(val) )

/*
** Reset segment registers
** with given segment selectors
*/
#define load_selectors(code, data)  asm volatile( "   ljmp   %0  , $1f   \n"\
                                                  "1:                    \n"\
                                                  "   movw   %1  , %%ax  \n"\
                                                  "   movw   %%ax, %%ds  \n"\
                                                  "   movw   %%ax, %%es  \n"\
                                                  "   movw   %%ax, %%fs  \n"\
                                                  "   movw   %%ax, %%gs  \n"\
                                                  "   movw   %%ax, %%ss  \n"\
                                                  ::"i"(code),"i"(data):"ax")



#endif
