#ifndef __INTERRUPTS_H__
#define __INTERRUPTS_H__

#include <types.h>
#include <segmentation.h>

/*
** Pmode IDT size
*/
#define IDT_SIZE                256

/*
** Gate types 
*/
#define INTERRUPT_GATE_32_TYPE  0xe  /* 1110 */
#define INTERRUPT_GATE_16_TYPE  0x6  /* 0110 */

#define TRAP_GATE_32_TYPE       0xf  /* 1111 */
#define TRAP_GATE_16_TYPE       0x7  /* 0111 */

#define TASK_GATE_TYPE          0x5  /* 0101 */

/*
** Real-mode IVT entry
*/
typedef struct ivt_entry
{
   uint16_t ip;
   uint16_t cs;

} __attribute__((packed)) ivt_entry_t;

/*
** IDT trampoline entry
*/
#define IDT_TRAMPOLINE_ALIGN    16

typedef struct idt_trampoline
{
   uint8_t bytes[IDT_TRAMPOLINE_ALIGN];

} __attribute__((packed)) idt_jmp_t;

/*
** Some BIOS interrupts
*/
#define BIOS_VIDEO_INTERRUPT          0x10
#define BIOS_DISK_INTERRUPT           0x13
#define BIOS_MISC_INTERRUPT           0x15
#define BIOS_BOOT_INTERRUPT           0x19

#define BIOS_NEW_GET_EXT_MEM          0xe801
#define BIOS_GET_EXT_MEM              0x8800


/*
** An interrupt descriptor entry:
**
** - interrupt gate (16 or 32)
** - trap gate (16 or 32)
** - task gate
*/
typedef struct interrupt_descriptor
{
   uint16_t  offset_1;    /* bits 00-15 of the isr offset */
   uint16_t  selector;    /* isr segment selector */
   uint_t    reserved:5;  /* intel reserved */
   uint_t    zero_1:3;    /* if not task gate should be 0 */
   uint_t    type:4;      /* gate type */
   uint_t    zero_2:1;    /* must be zero */
   uint_t    dpl:2;       /* desc privilege */
   uint_t    p:1;         /* present flag */
   uint16_t  offset_2;    /* bits 16-31 of the isr offset */

}__attribute__((packed,aligned(8))) interrupt_descriptor_t;

/*
** Prepare interrupt gate descriptor in "entry"
*/
#define set_interrupt_gate(entry,isr)			\
({                                                      \
   if( ! (isr) )					\
      (entry).p = 0;					\
   else							\
   {							\
      (entry).offset_1 = (isr)&0xffff;			\
      (entry).offset_2 = ((isr)>>16)&0xffff;		\
      (entry).selector = (K_CODE_SELECTOR);		\
      (entry).zero_1 = 0;				\
      (entry).zero_2 = 0;				\
      (entry).type = INTERRUPT_GATE_32_TYPE;		\
      (entry).dpl = 0;					\
      (entry).p = 1;					\
   }							\
})

/*
** Interrupt descriptor table register
*/
typedef struct interrupt_descriptor_table_register
{
   uint16_t   limit;      /* idt limit = 8*N-1 */
   uint32_t   base_addr;  /* idt addr */

}__attribute__((packed)) idt_reg_t;


/*
** Functions
*/
struct mem_map;

typedef void (*idt_handler_t)(ctx_t*);

void interrupts_init(struct mem_map*);
void interrupts_set_isr(interrupt_descriptor_t*, uint8_t, idt_handler_t);


#endif
