#ifndef __KBD_H__
#define __KBD_H__

#include <types.h>

/*
** Keyboard controler ports
*/
#define KBD_DATA_PORT              0x60
#define KBD_CTRL_PORT              0x64

/*
** Keyboard Controler commands : out(CTRL)
*/
#define KBD_CMD_READ_CMD_BYTE      0x20
#define KBD_CMD_WRITE_CMD_BYTE     0x60

#define KBD_CMD_SELF_TEST          0xaa
#define KBD_CMD_INTERFACE_TEST     0xab

#define KBD_CMD_DISABLE_AUX        0xa7
#define KBD_CMD_ENABLE_AUX         0xa8
#define KBD_CMD_TEST_AUX           0xa9

#define KBD_CMD_DISABLE_KBD        0xad
#define KBD_CMD_ENABLE_KBD         0xae

#define KBD_CMD_READ_I_PORT        0xc0

#define KBD_CMD_READ_O_PORT        0xd0
#define KBD_CMD_WRITE_O_PORT       0xd1

#define KBD_CMD_WRITE_AUX_O_PORT   0xd3

#define KBD_CMD_READ_TEST_INPUT    0xe0

#define KBD_CMD_SYSTEM_RESET       0xfe

/*
** Keyboard commands : out(DATA)
*/
#define KBD_CMD_LED_OFF            0x00
#define KBD_CMD_LED                0xed

#define KBD_CMD_ECHO               0xee
#define KBD_CMD_RESET              0xff

#define KBD_CMD_ENABLE_SCAN        0xf4
#define KBD_CMD_DISABLE_SCAN       0xf5
#define KBD_CMD_SCANCODE_SET       0xf0

#define KBD_CMD_PS2_READ_ID        0xf2

/*
** Commands results
*/
#define KBD_ANSWER_INTERFACE_TEST_OK          0x00
#define KBD_ANSWER_INTERFACE_TEST_CLOCK_LOW   0x01
#define KBD_ANSWER_INTERFACE_TEST_CLOCK_HIGH  0x02
#define KBD_ANSWER_INTERFACE_TEST_DATA_LOW    0x03
#define KBD_ANSWER_INTERFACE_TEST_DATA_HIGH   0x04

#define KBD_ANSWER_ACK                        0xfa

#define KBD_ANSWER_SELF_TEST_OK               0x55
#define KBD_ANSWER_SELF_TEST_KO               0xfc

#define KBD_ANSWER_ECHO                       0xee
#define KBD_ANSWER_ERROR                      0xff

#define KBD_ANSWER_BAT                        0xaa
#define KBD_ANSWER_BAT_FAILED                 0xfc

#define KBD_ANSWER_TOOMUCHKEY                 0x00

#define KBD_ANSWER_STD_PS2_ID                 0x00

/*
** LED parameters
*/
#define KBD_WRITE_LED_SCROLL_ON    (1<<0)
#define KBD_WRITE_LED_NUM_ON       (1<<1)
#define KBD_WRITE_LED_CAPS_ON      (1<<2)

/*
** Scancode set parameters
*/
#define KBD_WRITE_SCANCODE_SET_1      0x1
#define KBD_WRITE_SCANCODE_SET_2      0x2
#define KBD_WRITE_SCANCODE_SET_3      0x3
#define KBD_WRITE_SCANCODE_GET        0x0
#define KBD_READ_SCANCODE_1           'C'
#define KBD_READ_SCANCODE_2           'A'
#define KBD_READ_SCANCODE_3           '?'


/*  -- The KBD Rate --
**
** The 2nd byte is:
**   - bits 0-4 rate
**   - bits 5 & 6 pause before repeat
**   - bit 7 always 0
*/
#define KBD_WRITE_RATE               0xf3
/* 30 keys/sec, 26 ... */
#define KBD_WRITE_RATE_30            0x00
#define KBD_WRITE_RATE_26            0x01
#define KBD_WRITE_RATE_24            0x02
#define KBD_WRITE_RATE_20            0x04
#define KBD_WRITE_RATE_15            0x08
#define KBD_WRITE_RATE_10            0x0a
#define KBD_WRITE_RATE_9             0x0d
#define KBD_WRITE_RATE_7             0x10
#define KBD_WRITE_RATE_5             0x14
#define KBD_WRITE_RATE_2             0x1f
/* 250ms, 500ms, ... */
#define KBD_WRITE_PAUSE_250          0x00
#define KBD_WRITE_PAUSE_500          0x01
#define KBD_WRITE_PAUSE_750          0x02
#define KBD_WRITE_PAUSE_1000         0x04


/*
** Status register bits
*/
#define KBD_STATUS_OBF    (1<<0)         /* output buffer (0:empty) can not read  on 0x60 (1:full) can     read  on 0x60 */
#define KBD_STATUS_IBF    (1<<1)         /* input  buffer (0:empty) can     write to 0x60 (1:full) can not write to 0x60 */
#define KBD_STATUS_SYS    (1<<2)         /* (0) system in power-on reset, (1) system already initialized */
#define KBD_STATUS_A2     (1<<3)         /* (1) if out(CTRL) (0) if out(DATA) */
#define KBD_STATUS_INH    (1<<4)         /* (0) keyboard inhibited (1) keyboard is not inhibited */
#define KBD_STATUS_MOBF   (1<<5)         /* Like OBF but for ps/2 mouse */
#define KBD_STATUS_TO     (1<<6)         /* (0) no timeout error, (1) send or receive timeout */
#define KBD_STATUS_PERR   (1<<7)         /* (1) parity error */

typedef struct keyboard_status_register_fields
{
   uint_t  obf:1;
   uint_t  ibf:1;
   uint_t  sys:1;
   uint_t  cmd:1;
   uint_t  inh:1;
   uint_t  mobf:1;
   uint_t  to:1;
   uint_t  perr:1;

} __attribute__((packed)) kbd_status_reg_fields_t;

typedef struct keyboard_status_register
{
   union
   {
      uint8_t                   raw;
      kbd_status_reg_fields_t   fields;
   };

} __attribute__((packed)) kbd_status_reg_t;

/*
** Command byte
*/
#define KBD_CMD_BYTE_INT   (1<<0)       /* (1) irq1  is generated if OBF=1 (0) polling OBF needed */
#define KBD_CMD_BYTE_INT2  (1<<1)       /* (1) irq12 is generated if mouse present */
#define KBD_CMD_BYTE_SYS   (1<<2)       /* (1) set SYS bit in status register (0) clear SYS bit in status register */
#define KBD_CMD_BYTE_RES1  (1<<3)       /* unused on PS/2, on AT it is inhibit switch  */
#define KBD_CMD_BYTE_EN    (1<<4)       /* (0) keyboard enabled (1) keyboard disabled */
#define KBD_CMD_BYTE_EN2   (1<<5)       /* (0) mouse  enabled (1) mouse disabled */
#define KBD_CMD_BYTE_XLAT  (1<<6)       /* (0) disable translation of scancode */
#define KBD_CMD_BYTE_RES2  (1<<7)       /* unused on AT and PS/2 */

typedef struct keyboard_cmd_byte_fields
{
   uint_t  intr:1;
   uint_t  intr2:1;
   uint_t  sys:1;
   uint_t  r1:1;
   uint_t  _en:1;
   uint_t  _en2:1;
   uint_t  xlat:1;
   uint_t  r2:1;

} __attribute__((packed)) kbd_cmd_byte_fields_t;

typedef struct keyboard_cmd_byte
{
   union
   {
      uint8_t                   raw;
      kbd_cmd_byte_fields_t     fields;
   };

} __attribute__((packed)) kbd_cmd_byte_t;

/*
** Input Port
*/
#define KBD_INPUT_PORT_EXT_RAM      (1<<4)  /* (1) enable 2nd 256K of motherboard RAM */
#define KBD_INPUT_PORT_JUMPER       (1<<5)  /* (1) manufacturing jumper not installer */
#define KBD_INPUT_PORT_DISPLAY      (1<<6)  /* (1) MDA (0) CGA */
#define KBD_INPUT_PORT_NO_INHIBIT   (1<<7)  /* (1) kbd NOT inhibited */

typedef struct keyboard_input_port_fields
{
   uint_t  undef:4;
   uint_t  ext_ram:1;
   uint_t  jmp:1;
   uint_t  display:1;
   uint_t  n_inhib:1;

} __attribute__((packed)) kbd_input_port_fields_t;

typedef struct keyboard_input_port
{
   union
   {
      uint8_t                   raw;
      kbd_input_port_fields_t   fields;
   };

} __attribute__((packed)) kbd_input_port_t;

/*
** Output Port
*/
#define KBD_OUTPUT_PORT_SYS_RST      (1<<0)  /* System reset line */
#define KBD_OUTPUT_PORT_A20          (1<<1)  /* Gate A20 */
#define KBD_OUTPUT_PORT_OBF          (1<<4)  /* Output Buffer Full */
#define KBD_OUTPUT_PORT_IBE          (1<<5)  /* Input Buffer Empty */
#define KBD_OUTPUT_PORT_CLOCK        (1<<6)  /* Kbd Clock (output) */
#define KBD_OUTPUT_PORT_DATA         (1<<7)  /* Kbd Data (output) */

typedef struct keyboard_output_port_fields
{
   uint_t  sys_rst:1;
   uint_t  a20:1;
   uint_t  undef:2;
   uint_t  obf:1;
   uint_t  ibe:1;
   uint_t  clock:1;
   uint_t  data:1;

} __attribute__((packed)) kbd_output_port_fields_t;

typedef struct keyboard_output_port
{
   union
   {
      uint8_t                   raw;
      kbd_output_port_fields_t  fields;
   };

} __attribute__((packed)) kbd_output_port_t;


/*
** Reading/Writing data
*/
#define kbd_can_read_data()    (    in(KBD_CTRL_PORT) & KBD_STATUS_OBF )
#define kbd_can_write_data()   ( ! (in(KBD_CTRL_PORT) & KBD_STATUS_IBF) )

#define kbd_wait_read_data()   while( ! kbd_can_read_data()  );
#define kbd_wait_write_data()  while( ! kbd_can_write_data() );

#define __kbd_read_data()      in(KBD_DATA_PORT)
#define __kbd_write_data(_X_)  out((_X_),KBD_DATA_PORT)
#define __kbd_write_cmd(_X)    out((_X),KBD_CTRL_PORT)

#define kbd_read_status()      in(KBD_CTRL_PORT)
#define kbd_read_data()        ({ kbd_wait_read_data();  __kbd_read_data(); })
#define kbd_write_data(_x_)    ({ kbd_wait_write_data(); __kbd_write_data((_x_)); })

#define kbd_flush()            ({ while( kbd_can_read_data() ) __kbd_read_data(); })


/*
** Command are sent to CTRL PORT
** Command parameters are sent to DATA PORT
** Results are read from DATA PORT
**
** Even if OBF is for DATA PORT,
** if we want to write on CTRL PORT
** we have to check OBF
*/
#define PIT_SYS_CTRL_PORT               0x61

#define kbd_write_cmd(cmd)              ({ kbd_wait_write_data();           __kbd_write_cmd((cmd));  })
#define kbd_write_cmd_answ(cmd)         ({ kbd_write_cmd((cmd));              kbd_read_data();       })
#define kbd_write_cmd_prm(cmd,prm)      ({ kbd_write_cmd((cmd));            __kbd_write_data((prm)); })
#define kbd_write_cmd_prm_answ(cmd,prm) ({ kbd_write_cmd_prm((cmd),(prm));    kbd_read_data();       })

#define kbd_ack_scancode()              ({uint8_t x=in(PIT_SYS_CTRL_PORT); out(x|0x80,PIT_SYS_CTRL_PORT); out(x,PIT_SYS_CTRL_PORT);})

#define kbd_is_break_code(data)         (10000000&(data))
#define kbd_is_special_key(key)         ((scancode_std_table[ key ]>=KBD_BEGIN_SKEY)?1:0)



/*
** ASCII codes
*/
#define KBD_NB_SCANCODES   0x59
#define KBD_BEGIN_SKEY     1000
#define EXTENDED_CODE      0xe0

/*
** Special keys are never placed into the keyboard buffer
**
** other keys get there ascii code
** that will be converted from "int" to "char"
*/
/* bufferable keys */
#define KBD_UNUSED       0
#define KBD_PUP          1
#define KBD_HOME         2
#define KBD_END          3
#define KBD_PDW          4
#define KBD_BKSP         8
#define KBD_TAB          9
#define KBD_ENTER       10
#define KBD_SPACE       32
#define KBD_ESCAPE      27
#define KBD_DEL        127

/* non bufferable special keys */
#define KBD_CTRL_L    1000
#define KBD_SHIFT_L   1001
#define KBD_SHIFT_R   1002
#define KBD_ALT       1003

/* scan = 3a ba, mis a l'index 3a */
#define KBD_CAPS      1004

#define KBD_F1        1010
#define KBD_F2        1011
#define KBD_F3        1012
#define KBD_F4        1013
#define KBD_F5        1014
#define KBD_F6        1015
#define KBD_F7        1016
#define KBD_F8        1017
#define KBD_F9        1018
#define KBD_F10       1019
/* scan 57 et 58 */
#define KBD_F11       1020
#define KBD_F12       1021

/* scan = 45 c5, mis a l'index 45 */
#define KBD_NUM       1022

/* scan e047, mis a l'index 47 */
#define KBD_UP        1023
#define KBD_LEFT      1024
#define KBD_RIGHT     1025
#define KBD_DW        1026
#define KBD_INSERT    1027


/*
** Keyboard object
*/
#define KBD_MAX_BUFFER_SIZE  8

typedef struct keyboard_info
{
   uint_t                      *scancode_table;             /* the active scancode table */
   uint8_t                     buffer[KBD_MAX_BUFFER_SIZE]; /* kbd circular input buffer */
   uint_t                      b_h, b_q;                    /* buffer head & queue */
   uint8_t                     extended;                    /* scancode extension, 0xe0 */
   uint8_t                     shift_on;                    /* Some controls */
   uint8_t                     altgr_on;
   uint8_t                     ctrl_on;
   uint8_t                     alt_on;

}keyboard_info_t;

uint8_t  scancode_to_ascii(uint8_t);
void     process_special_key(uint8_t, uint8_t);
uint8_t  read_buffer();
void     fill_buffer(const uint8_t);

/*
** Kbd functions
*/
void      kbd_init();
void      kbd_isr(ctx_t*);
uint8_t   kbd_getkey();


#endif
