#ifndef __PIC_H__
#define __PIC_H__

#include <types.h>


/*
** Intel 8259a
*/

#define PIC1                  0x20
#define PIC2                  0xa0

/* A0 line = 0 */
#define PIC_ICW1(BASE)       (BASE)     /*  W  */
#define PIC_OCW2(BASE)       (BASE)     /*  W  */
#define PIC_OCW3(BASE)       (BASE)     /*  W  */
#define PIC_IRR(BASE)        (BASE)     /*  R  */
#define PIC_ISR(BASE)        (BASE)     /*  R  */

#define is_pic_icw1(_dAtA_)  ((_dAtA_)&0x10)
#define is_pic_ocw2(_dAtA_)  (((_dAtA_)&0x18) == 0 )
#define is_pic_ocw3(_dAtA_)  (((_dAtA_)&0x98) == 8 )

#define PIC_EOI               0x20      /* OCW2 non specific EOI */


/* A0 line = 1 */
#define PIC_IMR(BASE)       ((BASE)+1)  /* R/W */
#define PIC_ICW2(BASE)      ((BASE)+1)  /*  W  */
#define PIC_ICW3(BASE)      ((BASE)+1)  /*  W  */
#define PIC_ICW4(BASE)      ((BASE)+1)  /*  W  */


/*
** PIC Initialization command word 1
*/
typedef struct pic_init_command_word_1_fields
{
   uint_t   icw4:1;    /* (1) need icw4 */
   uint_t   single:1;  /* (1) single PIC (0) cascaded */
   uint_t   unused:1;  /* unused on x86 */
   uint_t   ltim:1;    /* (1) level triggered interrupt mode (0) edge triggered interrupt mode */
   uint_t   one:1;     /* MUST BE SET TO 1 */
   uint_t   vect:3;    /* Interrupt Vector Addresses for MCS mode */

} __attribute__((packed)) pic_icw1_fields_t;

typedef struct pic_init_command_word_1
{
   union
   {
      uint8_t            raw;
      pic_icw1_fields_t  fields;
   };

} __attribute__((packed)) pic_icw1_t;

/*
** PIC Initialization command word 2
**
** Base IRQ number for PIC1 and PIC2
** Each IRQ delivered by each PIC
** will be added to this base to form
** the final IRQ number
**
** For PIC 1:
** In real mode: IRQ0 is at IVT[8]
** In prot mode: IRQ0 is at IDT[32]
**
** The 3 less significant bits are
** always zero when writing ICW2.
** They correspond to irq number (0-7)
** related to the PIC
**
** This allow processor to associate
** IRQ with processor interrupt handler
**
*/
typedef struct pic_init_command_word_2_fields
{
   uint_t  zero:3;
   uint_t  irq_base:5;

} __attribute__((packed)) pic_icw2_fields_t;

typedef struct pic_init_command_word_2
{
   union
   {
      uint8_t            raw;
      pic_icw2_fields_t  fields;
   };

} __attribute__((packed)) pic_icw2_t;

/*
** PIC Initialization command word 3
**
** Slave connection
*/
typedef struct pic_init_command_word_3_master_fields
{
   uint_t   irq0:1;     /* irq0 is connected to a slave */
   uint_t   irq1:1;
   uint_t   irq2:1;
   uint_t   irq3:1;
   uint_t   irq4:1;
   uint_t   irq5:1;
   uint_t   irq6:1;
   uint_t   irq7:1;

} __attribute__((packed)) pic_icw3_m_fields_t;

typedef struct pic_init_command_word_3_slave_fields
{
   uint_t   slave_id:3; /* master irq number connected to slave */
   uint_t   zero:5;

} __attribute__((packed)) pic_icw3_s_fields_t;

typedef struct pic_init_command_word_3
{
   union
   {
      uint8_t              raw;
      pic_icw3_m_fields_t  m_fields;
      pic_icw3_s_fields_t  s_fields;
   };

} __attribute__((packed)) pic_icw3_t;

/*
** PIC Initialization command word 4
*/
typedef struct pic_init_command_word_4_fields
{
   uint_t   x86:1;    /* (1) 8086  (0) MCS */
   uint_t   aeoi:1;   /* (1) Auto EOI  (0) Normal EOI */
   uint_t   ms:1;     /* (1) master buffered (0) slave buffered */
   uint_t   buff:1;   /* (1) buffered mode (0) non buffered mode */
   uint_t   sfnm:1;   /* (1) Special Fully Nested Mode (0) Fully Nested Mode */
   uint_t   zero:3;

} __attribute__((packed)) pic_icw4_fields_t;

typedef struct pic_init_command_word_4
{
   union
   {
      uint8_t            raw;
      pic_icw4_fields_t  fields;
   };

} __attribute__((packed)) pic_icw4_t;

/*
** PIC Operation control word 1
**
** Control Interrupt Mask Register
**
** (0) irq active
** (1) irq inactive
*/
typedef struct pic_operation_control_word_1_fields
{
   uint_t  irq0:1;
   uint_t  irq1:1;
   uint_t  irq2:1;
   uint_t  irq3:1;
   uint_t  irq4:1;
   uint_t  irq5:1;
   uint_t  irq6:1;
   uint_t  irq7:1;

} __attribute__((packed)) pic_ocw1_fields_t;

typedef struct pic_operation_control_word_1
{
   union
   {
      uint8_t            raw;
      pic_ocw1_fields_t  fields;
   };

} __attribute__((packed)) pic_ocw1_t;

/*
** PIC Operation control word 2
*/
#define PIC_OCW2_DISABLE_ROTATE_AEOI       0
#define PIC_OCW2_NON_SPECIFIC_EOI          1
#define PIC_OCW2_NO_OPERATION              2
#define PIC_OCW2_SPECIFIC_EOI              3
#define PIC_OCW2_ENABLE_ROTATE_AEOI        4
#define PIC_OCW2_ROTATE_NON_SPECIFIC_EOI   5
#define PIC_OCW2_SET_PRIORITY              6
#define PIC_OCW2_ROTATE_SPECIFIC_EOI       7

typedef struct pic_operation_control_word_2_command_fields
{
   uint_t   eoi:1;
   uint_t   sl:1;
   uint_t   r:1;

} __attribute__((packed)) pic_ocw2_cmd_fields_t;

typedef struct pic_operation_control_word_2_command
{
   union
   {
      uint8_t                raw;
      pic_ocw2_cmd_fields_t  fields;
   };

} __attribute__((packed)) pic_ocw2_cmd_t;

typedef struct pic_operation_control_word_2_fields
{
   uint_t          lvl:3;   /* interrupt level when SL is set */
   uint_t          zero:2;
   uint_t          cmd:3;

} __attribute__((packed)) pic_ocw2_fields_t;

typedef struct pic_operation_control_word_2
{
   union
   {
      uint8_t            raw;
      pic_ocw2_fields_t  fields;
   };

} __attribute__((packed)) pic_ocw2_t;

/*
** PIC Operation control word 3
**
** Read Status Registers
*/
#define PIC_OCW3_SPECIAL_MASK_MODE_ENABLED   3
#define PIC_OCW3_SPECIAL_MASK_MODE_DISABLED  2

typedef struct pic_operation_control_word_3_fields
{
   uint_t   reg:2;   /* (00) reserved
		     ** (01) reserved
		     ** (10) Next Read on Base Register returns Interrupt Request Register
		     ** (11) Next Read on Base Register returns In-Service Register
		     */

   uint_t   poll:1;  /* (1) Poll command, (0) No poll command */
   uint_t   one:1;   /* MUST BE SET TO 1 */
   uint_t   zero:1;
   uint_t   smm:2;   /* (00) reserved
		     ** (01) reserved
		     ** (10) Reset Special Mask Mode
		     ** (11) Set Special Mask Mode
		     */
   uint_t   zero2:1;

} __attribute__((packed)) pic_ocw3_fields_t;

typedef struct pic_operation_control_word_3
{
   union
   {
      uint8_t            raw;
      pic_ocw3_fields_t  fields;
   };

} __attribute__((packed)) pic_ocw3_t;

/*
** Functions
*/
void   pic_init();

#define enable_clock()    out( in(PIC_IMR(PIC1))&(~1), PIC_IMR(PIC1) )
#define disable_clock()   out( in(PIC_IMR(PIC1))|1, PIC_IMR(PIC1) )

#endif

