#include <kbd.h>
#include <klibc.h>
#include <asmutils.h>

static keyboard_info_t kbd;

static uint_t scancode_std_table[] = {
   KBD_UNUSED, KBD_ESCAPE, '&', 130, '\"', '\'', '(', '-', 138, '_', 135, 133, ')', '=',
   KBD_BKSP, KBD_TAB, 'a', 'z', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '^', '$', KBD_ENTER,
   KBD_CTRL_L, 'q', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'm', 151, 253, KBD_SHIFT_L, '*',
   'w', 'x','c', 'v', 'b', 'n', ',', ';', ':', '!', KBD_SHIFT_R, KBD_UNUSED, KBD_ALT, KBD_SPACE,
   KBD_CAPS, KBD_F1, KBD_F2, KBD_F3, KBD_F4, KBD_F5, KBD_F6, KBD_F7, KBD_F8, KBD_F9, KBD_F10,
   KBD_NUM, KBD_UNUSED, KBD_HOME, KBD_UP, KBD_PUP, KBD_UNUSED, KBD_LEFT, KBD_UNUSED, KBD_RIGHT,
   KBD_UNUSED, KBD_END, KBD_DW, KBD_PDW, KBD_INSERT, KBD_DEL, KBD_UNUSED, KBD_UNUSED, '<',
   KBD_F11, KBD_F12
};

static uint_t scancode_shift_table[] = {
   KBD_UNUSED, KBD_ESCAPE, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 248, '+',
   KBD_BKSP, KBD_TAB, 'A', 'Z', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '^', 156, KBD_ENTER,
   KBD_CTRL_L, 'Q', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'M', '%', 253, KBD_SHIFT_L, 230,
   'W', 'X','C', 'V', 'B', 'N', '?', '.', '/', '', KBD_SHIFT_R, KBD_UNUSED, KBD_ALT, KBD_SPACE,
   KBD_CAPS, KBD_F1, KBD_F2, KBD_F3, KBD_F4, KBD_F5, KBD_F6, KBD_F7, KBD_F8, KBD_F9, KBD_F10,
   KBD_NUM, KBD_UNUSED, KBD_HOME, KBD_UP, KBD_PUP, KBD_UNUSED, KBD_LEFT, KBD_UNUSED, KBD_RIGHT,
   KBD_UNUSED, KBD_END, KBD_DW, KBD_PDW, KBD_INSERT, KBD_DEL, KBD_UNUSED, KBD_UNUSED, '>',
   KBD_F11, KBD_F12
};

static uint_t scancode_altgr_table[] = {
   KBD_UNUSED, KBD_ESCAPE, '', '~', '#', '{', '[', '|', '`', '\\', '^', '@', ']', '}',
   KBD_BKSP, KBD_TAB, 'a', 'z', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '^', '$', KBD_ENTER,
   KBD_CTRL_L, 'q', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'm', '', 253, KBD_SHIFT_L, '*',
   'w', 'x','c', 'v', 'b', 'n', ',', ';', ':', '!', KBD_SHIFT_R, KBD_UNUSED, KBD_ALT, KBD_SPACE,
   KBD_CAPS, KBD_F1, KBD_F2, KBD_F3, KBD_F4, KBD_F5, KBD_F6, KBD_F7, KBD_F8, KBD_F9, KBD_F10,
   KBD_NUM, KBD_UNUSED, KBD_HOME, KBD_UP, KBD_PUP, KBD_UNUSED, KBD_LEFT, KBD_UNUSED, KBD_RIGHT,
   KBD_UNUSED, KBD_END, KBD_DW, KBD_PDW, KBD_INSERT, KBD_DEL, KBD_UNUSED, KBD_UNUSED, '<',
   KBD_F11, KBD_F12
};


void kbd_init()
{
   kbd_cmd_byte_t    cmd_byte;
   kbd_input_port_t  input_port;
   kbd_status_reg_t  status;
   uint8_t           answ;

   memset( &kbd, 0, sizeof( keyboard_info_t ) );

   /* default scancode table */
   kbd.scancode_table = &scancode_std_table[0];

   /* read pending data before anything */
   kbd_flush();

   /* keyboard self test */
   answ = kbd_write_cmd_answ( KBD_CMD_SELF_TEST );
   answ = kbd_write_cmd_answ( KBD_CMD_INTERFACE_TEST );
   kbd_write_cmd( KBD_CMD_ENABLE_KBD );

   input_port.raw = kbd_write_cmd_answ( KBD_CMD_READ_O_PORT );
   if( ! (input_port.fields.ext_ram || input_port.fields.n_inhib) )
   {
      input_port.fields.ext_ram = 1;
      input_port.fields.n_inhib = 1;
      kbd_write_cmd_prm( KBD_CMD_WRITE_O_PORT, input_port.raw );
   }

   answ = kbd_write_cmd_answ( KBD_CMD_READ_O_PORT );

   /* read status register */
   status.raw = kbd_read_status();

   /* check command byte */
   cmd_byte.raw = kbd_write_cmd_answ( KBD_CMD_READ_CMD_BYTE );

   if( ! cmd_byte.fields.intr )
   {
      /* raise IRQ1 when OBF set */
      cmd_byte.fields.intr = 1;
      kbd_write_cmd_prm( KBD_CMD_WRITE_CMD_BYTE, cmd_byte.raw );
   }
}

uint8_t scancode_to_ascii(uint8_t code)
{
   if( code < KBD_NB_SCANCODES )
   {
      if( kbd_is_special_key(code) )
         process_special_key( code, 1 );
      else
         return kbd.scancode_table[ code ];
   }

   return KBD_UNUSED;
}

void process_special_key(uint8_t k, uint8_t pressed)
{
   switch( scancode_std_table[ k ] )
   {
   case KBD_CTRL_L:
      kbd.ctrl_on = pressed;
      break;

   case KBD_SHIFT_L:
   case KBD_SHIFT_R:
      kbd.shift_on = pressed;
      if( pressed )
	 kbd.scancode_table = &scancode_shift_table[0];
      else
	 kbd.scancode_table = &scancode_std_table[0];
      break;

   case KBD_ALT:
      if( kbd.extended )
      {
	 kbd.altgr_on = pressed;
	 if( pressed )
	    kbd.scancode_table = &scancode_altgr_table[0];
	 else
	    kbd.scancode_table = &scancode_std_table[0];
      }
      else
	 kbd.alt_on = pressed;
      break;
   }
}

uint8_t read_buffer()
{
   uint8_t ret;

   if( ! kbd.buffer[ kbd.b_h ] )
      return 0;

   ret = kbd.buffer[ kbd.b_h ];
   kbd.buffer[ kbd.b_h ] = 0;
   kbd.b_h = (++kbd.b_h)%KBD_MAX_BUFFER_SIZE;

   return ret;
}

void fill_buffer(uint8_t c)
{
   if( (kbd.b_q == kbd.b_h) && kbd.buffer[ kbd.b_q ] )
      return;

   kbd.buffer[ kbd.b_q ] = c;
   kbd.b_q = (++kbd.b_q)%KBD_MAX_BUFFER_SIZE;
}

void kbd_isr(ctx_t *ctx)
{
   uint8_t      key;
   uint8_t      ascii;

   if( ! kbd_can_read_data() )
      return;

   key = kbd_read_data();
   kbd_ack_scancode();
   
   if( key == EXTENDED_CODE )
   {
      kbd.extended = 1;
      return;
   }

   if( kbd_is_break_code( key ) )
   {
      if( kbd_is_special_key( key - 0x80 ) )
         process_special_key( key - 0x80, 0 );
   }
   else
   {
      ascii = scancode_to_ascii( key );
      if( ascii )
         fill_buffer( ascii );
   }

   if( kbd.extended )
      kbd.extended = 0;
}

uint8_t kbd_getkey()
{
   uint8_t rc;

   while( ! (rc=read_buffer()) )
      asm volatile ( "hlt" );

   return rc;
}

